#coding: utf-8

from importlib import import_module
from multiprocessing import Process, Queue
from optparse import OptionParser
from Queue import Empty
from setproctitle import setproctitle

import os
import pystatsd
import yaml

from log import logger, setup_logging


def prepare_monitor(name, collector, config=None):
    def starter(*args, **kwargs):
        setproctitle('tipster: {0} monitor'.format(name))
        try:
            kwargs['config'] = config
            collector(*args, **kwargs)
        except KeyboardInterrupt:
            pass
    return starter

def start_monitors(feed_queue, configs=None, only=None):
    '''
    Запускаем подпроцессы для каждого из модулей мониторинга.
    '''
    if not configs:
        configs = {}

    modules_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'monitors')
    
    modules = [fn for fn in os.listdir(modules_path) if fn.endswith('.py')]

    if only:
        logger.debug('Attempting to start monitors: {0}'.format(', '.join(only)))
    else:
        logger.debug('Attempting to start all monitors.')
    processes = []
    for fn in modules:
        if fn == '__init__.py':
            continue

        name, ext = os.path.splitext(fn)
        if only and name not in only:
            logger.debug('Skipping "{0}"...'.format(name))
            continue

        module = import_module('.monitors.{0}'.format(name), package=__package__)
        logger.debug('Starting "{0}"...'.format(name))
        proc = Process(target=prepare_monitor(name, module.collect,
                                              config=configs.get(name)),
                       args=(feed_queue,))
        proc.start()
        logger.info('Monitor "{0}" started (PID: {1}).'.format(name, proc.pid))
    return processes


def main():
    setup_logging()

    cfg_files = ['./tipster.yaml']
    parser = OptionParser()
    parser.add_option('-c', '--cfg', dest='cfg_file',
                      help='Load additional config file.', metavar='<CONFIG>')
    parser.add_option('-d', '--dump', dest='just_dump', action='store_true',
                      help='Just dump results to stdout.')
    (options, args) = parser.parse_args()

    if options.cfg_file:
        cfg_files.append(options.cfg_file)

    config = {}
    for cfg_fn in cfg_files:
        logger.info('Loading configuration from "{0}".'.format(cfg_fn))
        try:
            with open(cfg_fn, 'r') as cfg_file:
                tmp_cfg = yaml.load(cfg_file)
                config.update(tmp_cfg)
        except Exception as se:
            logger.error("{0}: {1}".format(cfg_fn, se))

    identity = config.get('identity', {})
    
    client = identity.get('client')
    region = identity.get('region')
    product = identity.get('product')
    instance = identity.get('instance')

    statsd_conf = config.get('statsd', {})
    host = statsd_conf.get('host', 'localhost')    
    port = statsd_conf.get('statsd', 8125)

    # Даже если эти значения есть в конфиге, они все должны содержать в себе
    # какие-либо данные.    
    if not all([client, region, product]):
        logger.error('You must fill out [identity] section in config file!')
        return

    statsd = pystatsd.Client(host, port)
    metric_handlers = {
        'gauge': statsd.gauge,
        'incr': statsd.increment,
        'decr': statsd.decrement,
        'timing': statsd.timing
    }

    monitors_cfg = config.get('monitors')

    # Вычисляем протокольный префикс для отправки данных в БД мониторинга
    # (https://31.13.131.67/projects/dev-py/wiki/Monitoring-send-format)
    if instance:
        identity_prefix = '{0}.{1}.{2}.instance_{3}'.format(region, client, product, instance)
    else:
        identity_prefix = '{0}.{1}.{2}'.format(region, client, product)

    setproctitle('tipster: {0} -> {1}:{2}'.format(identity_prefix, host, port))
    logger.info('Our identity prefix: {0}'.format(identity_prefix))

    q = Queue()
    monitors = start_monitors(q, configs=monitors_cfg,
                                 only=config.get('start_only'))

    while True:        
        try:
            metric, value, m_type = q.get(True, 10)
            assert m_type in metric_handlers, 'Unknown metric type: "{0}"'.format(m_type)

            logger.debug('Received data: {0} {1} {2}'.format(metric, value, m_type))
                        
            if not options.just_dump:
                handler = metric_handlers[m_type]
                handler('.'.join(['v1', identity_prefix, metric]), value)
        except Empty:
            pass
        except KeyboardInterrupt:
            print '\nCtrl-C, Ctrl-V.'
            break

    logger.info('Shutting down...')

if __name__ == '__main__':
    main()
